                    (* WBLIB.PAS *)
Unit WBlib;

interface

uses B_cons;


type ConstnVec = array[1..startmax] of real;
     Matrix = array[1..startmax,1..startmax] of real;
     pVec = array[1..cmax] of real;
     pMatrix = array[1..cmax,1..cmax] of real;
     Barray = array[0..cmax] of Boolean;
     Sarray = array[1..cmax] of string;
     SIarray = array[1..cmax] of shortint;

function Root(x: real): real;

function log(x: real): real;

function logit(x:real):real;

function expo(x:real):real;

function expit(x:real):real;

function d(a,b: LongInt):real;

function divi(a,b: real): real;

function sign(a: boolean): integer;

function sgn(x: real): integer;

function square(a: integer): longint;
(* Note that sqr(a) goes wrong for bytes a >= 16 *)

(*procedure Outputfile(var fn1: String; fn2: String; var f: Text);
 Makes an outputfile f with filename fn1, for which fn2 is the default;
   The file gets a header with filename and date. *)

(*procedure Outputfiles(var fn1, fn2: String; fn3: String; var f1, f2: Text);
 Makes two outputfiles f with filename fn1, fn2;
   the defaults for fn1 and fn2 are fn3.OUT and fn3.PQR;
   the files get a header with filename and date. *)

(*procedure Outpfile(fn: String; var f: Text);
 Makes an outputfile f with filename fn;
   The file gets a header with filename and date. *)

procedure quicksort(var a: ConstnVec; Lo,Hi: integer);
(* sorts vector a, elements Lo through Hi, from small to high *)

procedure product(var a,b,c: Matrix);
(* c := a*b *)

{
procedure matpower(var a: matrix; p: integer; var b: matrix);
(* b := a**p *)
}
procedure pproduct(p: integer; var a,b,c: pMatrix);
(* c := a*b *)

procedure pprod(p: integer; var a: pmatrix; var b,c: pvec);
(* c := a*b as a px1 vector *)

function veclength(p: integer; var a: pvec; var s: pmatrix): real;
(* length := a'*inv(s)*a *)

procedure pmatpower(var a: pmatrix; q: integer; var b: pmatrix);
(* b := a**q *)

PROCEDURE pinverse(p: integer; var a: pMatrix; var y: pMatrix;
                   var errtext: string);
(* calculates matrix inverse: y := inverse(a),
   where dimension is taken to be p. *)

function power(a: real; b: integer): real;
(* returns a**b by repeated multiplication; use only for small b. *)

function max(a,b: real):real;
(* maximum of a and b *)

function st(i: integer):string;
(* The string containing the number i . *)

function one(a: boolean): integer;
(* 1 if a is true, 0 otherwise *)

function st2(r: real):string;
(* The string containing the number r . *)

var lf : textfile;     (* for writing log messages *)


implementation

uses SysUtils, Forms;

function Root(x: real): real;
begin
  if abs(x) > 1e-300 then Root := sqrt(abs(x)) else Root := 1e-150;
end; (* Root *)

function log(x: real): real;
begin
  if ((x > 1e-300) and (x < 1e300)) then
    log := ln(x)
   else if (x <= 1e-300) then
    log := ln(1e-300) else
    log := ln(1e300);
end;

function logit(x:real):real;
begin
  logit := log(x/(1.0-x))
end;

function expo(x:real):real;
begin
  if (x < -700.0) then expo := exp(-700.0) else
  begin if (x > 700.0) then expo := exp(700.0) else expo := exp(x) end;
(* In Delphi 6, the Real type ranges to wider than 1e-308 *)  
end;

function expit(x:real):real;
var y: real;
begin
  y := expo(x);
  expit := y/(1.0+y);
end;

function d(a,b: LongInt):real;
begin
  if (b >= 1) then d := a/b else d := 0.0;
end;

function divi(a,b: real): real;
begin
  if (abs(b) > 1e-12) then divi := a/b else divi := 0.0;
end;

function sign(a: boolean): integer;
begin
sign:=-1;
if a then sign:=1;
end;

function sgn(x: real): integer;
begin
sgn:=0;
if (x > 0) then sgn:=1 else
 if (x < 0) then sgn := -1;
end;

function square(a: integer): longint;
var aa: longint;
begin
  aa := a;
  square := sqr(aa);
end;
(*
procedure Outputfile(var fn1: String; fn2: String; var f: Text);
var  year, month, day, weekday, hour, min, sec, hund: word;
     fn0: String[20];
begin
fn0 := fn2;
fn1 := fn0;
assign(f,fn0);
rewrite(f);
writeln(f,'                        ****************** ');
writeln(f,'                         ',fn0);
writeln(f,'                        ****************** ');
writeln(f);
writeln(f);
writeln(f,'Filename is ',fn0,' .');
writeln(f);
writeln(f,'Time: '+DateTimeToStr(Now));
writeln(f); writeln(f);
writeln(f); writeln(f);
end;

procedure Outpfile(fn: String; var f: Text);
var  year, month, day, weekday, hour, min, sec, hund: word;
     fn0: String[20];
begin
writeln;
assign(f,fn);
rewrite(f);
writeln(f,'                        ****************** ');
writeln(f,'                         ',fn);
writeln(f,'                        ****************** ');
writeln(f);
writeln(f);
writeln(f,'Filename is ',fn,' .');
writeln(f);
writeln(f,'Time: '+DateTimeToStr(Now));
writeln(f); writeln(f);
writeln(f); writeln(f);
end;

procedure Outputfiles(var fn1, fn2: String; fn3: String; var f1, f2: Text);
var  year, month, day, weekday, hour, min, sec, hund: word;
     fn0: String; sa: string; sb: string; sc: string;
procedure OPF(fn: string; var f: Text);
 begin
 assign(f,fn);
 rewrite(f);
 writeln(f,'                        ****************** ');
 writeln(f,'                         ',fn);
 writeln(f,'                        ****************** ');
 writeln(f);
 writeln(f);
 writeln(f,'Time: '+DateTimeToStr(Now));
 writeln(f); writeln(f);
 writeln(f); writeln(f);
end;
begin
fn0 := fn3+'.OUT';
fn1 := fn0;
OPF(fn1,f1);
fn2:=changefileext(fn1,'.PQR');
OPF(fn2,f2);
writeln(f2,'Secondary output file for some extra details.');
writeln(f1,'Secondary output is on file ',fn2,'.');
end;
     *)

procedure quicksort(var a: ConstnVec; Lo,Hi: integer);

procedure sort(l,r: integer);
var
  i,j: integer;
  x,y: real;
begin
  i:=l; j:=r; x:=a[(l+r) DIV 2];
  repeat
    while a[i]<x do i:=i+1;
    while x<a[j] do j:=j-1;
    if i<=j then
    begin
      y:=a[i]; a[i]:=a[j]; a[j]:=y;
      i:=i+1; j:=j-1;
    end;
  until i>j;
  if l<j then sort(l,j);
  if i<r then sort(i,r);
end;

begin (* quicksort *);
  sort(Lo,Hi);
end;

procedure product(var a,b,c: Matrix);
var i,j,h: integer; temp: real;
begin
  for i:=1 to startmax do for j:=1 to startmax do begin
    temp := 0.0;
    for h:=1 to startmax do temp := temp + a[i,h]*b[h,j];
    c[i,j] := temp;
   end;
end; (* product *)
{
procedure matpower(var a: matrix; p: integer; var b: matrix);
var aa,bb,c: matrix; i: integer;
begin
  aa:= a; bb:= a;
  for i:=p downto 2 do begin
    product(aa,bb,c); aa:=c; end;
  b := c;
end;
}

PROCEDURE pinverse(p: integer; var a: pMatrix; var y: pMatrix;
                   var errtext: string);
(* berekent matrixinverse *)

type rivec   = ARRAY[1..cmax] OF integer;
     rvec    = ARRAY[1..cmax] OF real;

VAR i,j: integer;
    d: real;
    vecy: rvec;
    indx: rivec;
    aa,ee: pmatrix;

(* programs using inverse must define the constant cmax and the type
TYPE pMatrix = ARRAY[1..cmax,1..cmax] OF real;
   in the main program *)

PROCEDURE ludcmp(VAR a: pMatrix; n,np: integer;
       VAR indx: rivec; VAR d: real; VAR errtext: string);
  (*nodig voor matrixinverteren*)
CONST
   tiny=1.0e-15;
VAR
   k,j,imax,i: integer;
   som,dum,big: real;
   vv: rvec;
BEGIN
   errtext := '';
   d := 1.0;
   FOR i := 1 to n DO BEGIN
      big := 0.0;
      FOR j := 1 to n DO IF (abs(a[i,j]) > big) THEN big := abs(a[i,j]);
      IF (big = 0.0) THEN BEGIN
         errtext :='error in LUDCMP - singular matrix'; exit
      END;
      vv[i] := 1.0/big
   END;
   FOR j := 1 to n DO BEGIN
      IF (j > 1) THEN BEGIN
         FOR i := 1 to j-1 DO BEGIN
            som := a[i,j];
            IF (i > 1) THEN BEGIN
               FOR k := 1 to i-1 DO BEGIN
                  som := som-a[i,k]*a[k,j]
               END;
               a[i,j] := som
            END
         END
      END;
      big := 0.0;
      FOR i := j to n DO BEGIN
         som := a[i,j];
         IF (j > 1) THEN BEGIN
            FOR k := 1 to j-1 DO BEGIN
               som := som-a[i,k]*a[k,j]
            END;
            a[i,j] := som
         END;
         dum := vv[i]*abs(som);
         IF (dum > big) THEN BEGIN
            big := dum;
            imax := i
         END
      END;
      IF (j <> imax) THEN BEGIN
         FOR k := 1 to n DO BEGIN
            dum := a[imax,k];
            a[imax,k] := a[j,k];
            a[j,k] := dum
         END;
         d := -d;
         vv[imax] := vv[j]
      END;
      indx[j] := imax;
      IF (j <> n) THEN BEGIN
         IF (abs(a[j,j]) < tiny) THEN a[j,j] := tiny;
         dum := 1.0/a[j,j];
         FOR i := j+1 to n DO BEGIN
            a[i,j] := a[i,j]*dum
         END
      END
   END;
   IF (abs(a[n,n]) < tiny) THEN a[n,n] := tiny
END;

PROCEDURE lubksb(a: pMatrix; n,np: integer; indx: rivec; VAR b: rvec;
                 var errtext: string);
VAR
   j,ip,ii,i: integer;
   som: real;
BEGIN
   errtext := '';
   ii := 0;
   FOR i := 1 to n DO BEGIN
      ip := indx[i];
      som := b[ip];
      b[ip] := b[i];
      IF  (ii <> 0) THEN BEGIN
         FOR j := ii to i-1 DO BEGIN
            som := som-a[i,j]*b[j]
         END
      END ELSE IF (som <> 0.0) THEN BEGIN
         ii := i
      END;
      b[i] := som
   END;
   FOR i := n DOWNTO 1 DO BEGIN
      som := b[i];
      IF (i < n) THEN BEGIN
         FOR j := i+1 to n DO BEGIN
            som := som-a[i,j]*b[j]
         END
      END;
      if (abs(a[i,i]) < 1.0e-15) then begin
        errtext :='error in routine lubskb in BLIB: a[i,i] small.';
        exit;
       end;
      b[i] := som/a[i,i]
   END
END;

BEGIN (* inverse *)
errtext := '';
aa := a;
if (p <= 0) then errtext := 'dimension not positive';
if (p > cmax) then errtext := 'dimension larger than cmax';
if (errtext > '') then exit;
FOR i := 1 TO cmax Do begin
  FOR j := 1 TO cmax DO y[i,j] := 0;
  y[i,i] := 1;
 END;
ludcmp(aa,p,p,indx,d,errtext);
if (errtext > '') then exit;
FOR j := 1 TO p DO BEGIN
  FOR i := 1 TO p DO vecy[i]:=0;
  vecy[j] := 1;
  lubksb(aa,p,p,indx,vecy,errtext);
  if (errtext > '') then exit;
  FOR i:= 1 TO p DO y[i,j] := vecy[i];
 END;
(* pproduct(p,a,y,ee);
for i:=1 to p do for j:=1 to p do
 if (((i=j) and (abs(ee[i,j]-1.0) > 1e-8))
   or ((i <> j) and (abs(ee[i,j]) > 1e-8)))
    then errtext := 'Matrix inversion not succesful.'; *)
END; (* inverse *)

procedure pproduct(p: integer; var a,b,c: pMatrix);
var i,j,h: integer; temp: real;
begin
  for i:=p downto 1 do for j:=p downto 1 do begin
    temp := 0.0;
    for h:=p downto 1 do temp := temp + a[i,h]*b[h,j];
    c[i,j] := temp;
   end;
end; (* pproduct *)

procedure pprod(p: integer; var a: pmatrix; var b,c: pvec);
var i,h: integer; temp: real;
begin
  for i:=p downto 1 do begin
    temp := 0.0;
    for h:=p downto 1 do temp := temp + a[i,h]*b[h];
    c[i] := temp;
   end;
end; (* pprod *)

function veclength(p: integer; var a: pvec; var s: pmatrix): real;
var sinv: pmatrix; ec: string; temp: real; i,j: integer;
begin
  pinverse(p,s,sinv,ec);
  if (ec > '') then
    for i:=p downto 1 do begin
      for j:=p downto 1 do sinv[i,j] := 0.0; sinv[i,i] := 1; end;
  temp := 0.0;
  for i:=p downto 1 do for j:= p downto 1 do
    temp := temp+a[i]*a[j]*sinv[i,j];
  veclength := temp;
end;

procedure pmatpower(var a: pmatrix; q: integer; var b: pmatrix);
var aa,bb,c: pmatrix; i: integer;
begin
  aa:= a; bb:= a;
  for i:=q downto 2 do begin
    pproduct(q,aa,bb,c); aa:=c; end;
  b := c;
end;

function power(a: real; b: integer): real;
(* This is faster than working with the logarithm for b <= about 25 *)
var x,aa: real; i: integer;
begin
  x := 1.0; aa := a;
  if (b <= 25) then
    for i := b downto 1 do x := x*aa
   else
     x := expo(b*log(aa));
  power := x;
end;

function max(a,b: real):real;
begin
  if (a < b) then max := b else max := a;
end;

function one(a: boolean): integer;
begin
  if a then one := 1 else one := 0;
end;

function st(i: integer):string;
(* The string containing the number i . *)
var ss: string;
begin
  str(i,ss);
  st := ss;
end; (* st *)

function st2(r: real):string;
(* The string containing the number i . *)
var ss: string;
begin
  str(r:3:2,ss);
  st2 := ss;
end; (* st *)


begin
end.
